package cern.jira.emailhandler;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.mail.Address;
import javax.mail.Message;
import javax.mail.MessagingException;
import javax.mail.internet.InternetAddress;

import org.apache.log4j.Category;
import org.ofbiz.core.entity.GenericValue;

import com.atlassian.core.user.UserUtils;
import com.atlassian.jira.ComponentManager;
import com.atlassian.jira.ManagerFactory;
import com.atlassian.jira.config.ConstantsManager;
import com.atlassian.jira.issue.IssueFieldConstants;
import com.atlassian.jira.issue.IssueImpl;
import com.atlassian.jira.issue.MutableIssue;
import com.atlassian.jira.issue.fields.CustomField;
import com.atlassian.jira.issue.fields.SummarySystemField;
import com.atlassian.jira.issue.issuetype.IssueType;
import com.atlassian.jira.issue.security.IssueSecurityLevelManager;
import com.atlassian.jira.mail.MailThreadManager;
import com.atlassian.jira.project.ProjectManager;
import com.atlassian.jira.security.Permissions;
import com.atlassian.jira.service.util.handler.AbstractMessageHandler;
import com.atlassian.jira.util.EasyList;
import com.atlassian.jira.web.action.util.JiraLicenseUtils;
import com.atlassian.jira.web.bean.FieldVisibilityBean;
import com.atlassian.jira.workflow.WorkflowFunctionUtils;
import com.atlassian.license.License;
import com.atlassian.license.LicenseManager;
import com.atlassian.mail.MailUtils;
import com.opensymphony.user.EntityNotFoundException;
import com.opensymphony.user.User;
import com.opensymphony.util.TextUtils;


/**
 * A message handler to create a new issue from an incoming message. 
 * Note: requires public noarg constructor as this
 * class is instantiated by reflection.
 * 
 * @author <a href="http://draffo.web.cern.ch">Daniele Raffo</a> 
 * on the basis of the CreateIssueHandler class, copyright (c) 2002-2006 Atlassian
 * @version 2.3 (29-5-2007)
 */
public class AdvancedCreateIssueHandler extends AbstractMessageHandler {

	private static final Category log = Category.getInstance(AdvancedCreateIssueHandler.class);
    private IssueDescriptor issueDescriptor;
    private Collection priorities;
    public String projectKey = null;         // chosen project for the new issue
    public String issueType;                 // chosen issue type 
    public String defaultProjectKey = null;  // default project where new issues are created
    public String defaultIssueType = null;   // default type for new issues
    public boolean ccAssignee = true;        // first Cc'ed user becomes the assignee ?
    private static final String KEY_PROJECT = "project";
    private static final String KEY_ISSUETYPE = "issuetype";
    private static final String CC_ASSIGNEE = "ccassignee";

    
    public void init(Map params) {
        log.debug("CreateIssueHandler.init(params: " + params + ")");

        super.init(params);

        if (params.containsKey(KEY_PROJECT)) {
            defaultProjectKey = (String) params.get(KEY_PROJECT);
        }

        if (params.containsKey(KEY_ISSUETYPE)) {
            defaultIssueType = (String) params.get(KEY_ISSUETYPE);
        }

        if (params.containsKey(CC_ASSIGNEE)) {
            ccAssignee = Boolean.valueOf((String) params.get(CC_ASSIGNEE)).booleanValue();
        }
        
        log.debug("Params: " + defaultProjectKey + " - " + defaultIssueType + " - " + ccAssignee);
    }

    
    public boolean handleMessage(Message message) throws MessagingException {
        log.debug("AdvancedCreateIssueHandler.handleMessage");
        
        if (issueDescriptor.getIssueType() != null) {   
        	issueType = issueDescriptor.getIssueType();
        } else {
        	issueType = defaultIssueType;
        }
        
        if (!canHandleMessage(message)) {
            return deleteEmail;
        }

        try {
        	String reporterName = issueDescriptor.getReporter();
        	// Sets the reporter specified in the tag, or the sender of the message, 
        	// or the default reporter
        	User reporter = (
        			reporterName != null && getUserByName(reporterName) != null ?
        					getUserByName(reporterName) :
       						getReporter(message)	
        	);
            if (reporter == null) {
                String errorMessage = "Sender is anonymous, no default reporter specified and creating users " +
                        "is set to false (or external user managment is enabled). Message rejected.";
                log.warn(errorMessage);
                return addErrorAndReturnFalse(errorMessage);
            }

            GenericValue project = getProject();
            log.debug("Project = " + project);
            if (project == null) {
            	project = getProjectManager().getProjectByKey(defaultProjectKey.toUpperCase());
            }
            if (project == null) {
                String errorMessage = "Cannot handle message as destination project is null and no default project set";
                log.warn(errorMessage);
                return addErrorAndReturnFalse(errorMessage);
            }

            log.debug("Issue Type Key = = " + issueType);
            if (!hasValidIssueType()) {
                String errorMessage = "Cannot handle message as Issue Type is null or invalid";
                log.warn(errorMessage);
                return addErrorAndReturnFalse(errorMessage);
            }
            String summary = issueDescriptor.getSummary();   
            if (!TextUtils.stringSet(summary)) { // should never happen
                return addErrorAndReturnFalse("Issue must have a summary. The mail message has an empty or no subject.");
            }
            if (summary.length() > SummarySystemField.MAX_LEN.intValue()) {
                log.warn("Truncating summary field because it is too long: " + summary);
                summary = summary.substring(0, SummarySystemField.MAX_LEN.intValue() - 3) + "...";
            }

            // JRA-7646 - check if priority/description is hidden - if so, do not set
            String priority = null;
            String description = null;
            FieldVisibilityBean visibility = new FieldVisibilityBean();

            if (!visibility.isFieldHiddenInAllSchemes(project.getLong("id"), IssueFieldConstants.PRIORITY, EasyList.build(issueType))) {
                priority = getPriority();
            }

            if (!visibility.isFieldHiddenInAllSchemes(project.getLong("id"), IssueFieldConstants.DESCRIPTION, EasyList.build(issueType))) {
                description = getDescription(reporter, message);
            }

            MutableIssue issueObject = IssueImpl.getIssueObject(null);
            
            issueObject.setProject(project);
            issueObject.setSummary(summary);
            issueObject.setDescription(description);
            issueObject.setIssueTypeId(issueType);
            issueObject.setReporter(reporter);

            // Sets the assignee specified in the tag, or the first valid Cc'ed assignee, 
            // or else the default assignee
            String assigneeName = issueDescriptor.getAssignee();
            User assignee = null;
            if (
            		assigneeName != null && 
            		getUserByName(assigneeName) != null && 
            		isValidAssignee(project, getUserByName(assigneeName))) {
            	assignee = getUserByName(assigneeName);
            } else {
            	if (ccAssignee) {
            		assignee = getFirstValidAssignee(message.getAllRecipients(), project);
            	}
            }
            if (assignee == null) {
                assignee = ComponentManager.getInstance().getAssigneeResolver().getDefaultAssignee(issueObject, Collections.EMPTY_MAP);
            }
            if (assignee != null) {
                issueObject.setAssignee(assignee);
            }
            
            issueObject.setPriorityId(priority);
            
            // Sets components to the issue
            if (issueDescriptor.getComponents() != null) {   // The component name
            	String[] comps = issueDescriptor.getComponents();
        		Collection components = new ArrayList();
            	for (int i = 0; i < comps.length; i++) {
            		GenericValue component = getProjectManager().getComponent(project, comps[i]);
            		if (component != null) {
            			components.add(component);
            		}
            	}
            	if (components.size() > 0) {
            		issueObject.setComponents(components);
            	}
            }

            // Ensure issue level security is correct
            setDefaultSecurityLevel(issueObject);

            Map fields = new HashMap();
            fields.put("issue", issueObject);
            GenericValue originalIssueGV = ComponentManager.getInstance().getIssueManager().getIssue(issueObject.getId());

            // Give the CustomFields a chance to set their default values JRA-11762
            List customFieldObjects = ComponentManager.getInstance().getCustomFieldManager().getCustomFieldObjects(issueObject);
            for (Iterator iterator = customFieldObjects.iterator(); iterator.hasNext();) {
                CustomField customField = (CustomField) iterator.next();
                issueObject.setCustomFieldValue(customField,  customField.getDefaultValue(issueObject));
            }

            fields.put(WorkflowFunctionUtils.ORIGNAL_ISSUE_KEY, IssueImpl.getIssueObject(originalIssueGV));
            GenericValue issue = ManagerFactory.getIssueManager().createIssue(reporter, fields);

            if (issue != null) {
                log.info("Issue " + issue.get("key") + " created");

                // Record the message id of this e-mail message so we can track replies to this message
                // and associate them with this issue
                recordMessageId(MailThreadManager.ISSUE_CREATED_FROM_EMAIL, message, issue.getLong("id"));
            }

            createAttachmentsForMessage(message, issue);

            return true;
        }
        catch (Exception e) {
            String errorMessage = "Could not create issue!";
            log.error(errorMessage, e);
            addError(errorMessage, e);
        }

        // something went wrong - don't delete the message
        return false;
    }

    
    protected GenericValue getProject() {
    	String pkey = issueDescriptor.getProjectKey().toUpperCase();
    	if (
    			pkey != null && 
    			!pkey.equals("") &&
    			getProjectManager().getProjectByKey(pkey) != null) {
    		return getProjectManager().getProjectByKey(pkey);    	
    	} else {
    		return null;
    	}
    }

    
    protected boolean hasValidIssueType() {
        // if there is no project then the issue cannot be created
        if (issueType == null) {
            log.debug("Issue Type NOT set. Cannot find Issue type.");
            return false;
        }

        IssueType issueTypeObject = ManagerFactory.getConstantsManager().getIssueTypeObject(issueType);
        if (issueTypeObject == null) {
            log.debug("Issue Type does not exist with id of " + issueType);
            return false;
        }

        log.debug("Issue Type Object = " + issueTypeObject.getName());
        return true;
    }

    
    protected ProjectManager getProjectManager() {
        return ManagerFactory.getProjectManager();
    }

    
    /**
     * Extracts the description of the issue from the message.
     * 
     * @param reporter  the established reporter of the issue
     * @param message  the message from which the issue is created
     * @return the description of the issue
     * @throws MessagingException
     */
    private String getDescription(User reporter, Message message) throws MessagingException {
        return recordFromAddressForAnon(reporter, message, MailUtils.getBody(message));
    }

    
    /**
     * Adds the senders' from addresses to the end of the issue's details (if they could be extracted), if the e-mail
     * has been received from an unknown e-mail address and the mapping to an "anonymous" user has been enabled.
     * 
     * @param reporter  the established reporter of the issue (after one has been established)
     * @param message  the message that is used to create issue
     * @param description  the issues extracted description
     * @return the modified description if the e-mail is from anonymous user, unmodified description otherwise
     * @throws MessagingException
     */
    private String recordFromAddressForAnon(User reporter, Message message, String description) throws MessagingException {
        // If the message has been created for an anonymous user add the senders e-mail address to the description.
        if (reporteruserName != null && reporteruserName.equals(reporter.getName())) {
            description += "\n[Created via e-mail ";
            if (message.getFrom() != null && message.getFrom().length > 0) {
                description += "received from: " + message.getFrom()[0] + "]";
            } else {
                description += "but could not establish sender's address.]";
            }
        }
        return description;
    }

    
    /**
     * Get the issue's priority.
     */
    private String getPriority() {
        ConstantsManager constantsManager = ManagerFactory.getConstantsManager();

        int priorityValue = issueDescriptor.getPriority();

        if (priorityValue == 0) {
        	return getDefaultSystemPriority();
        }

        // if priority is unset - pick the closest priority, this should be a sensible default
        priorities = constantsManager.getPriorities();

        Iterator priorityIt = priorities.iterator();

        /*
         * NOTE: Valid values for priority are (1=Highest, 2=High, 3=Normal, 4=Low & 5=Lowest). 
         * The parsed priority is divided by 5 (number of valid values) this gives the percentage
         * representation of the priority. We multiply this by the priority.size() 
         * (number of priorities in jira) to scale and map the percentage to a priority in jira.
         */
        int priorityNumber = (int) Math.ceil(((double) priorityValue / 5d) * (double) priorities.size());
        // if priority is too large, assume its the 'lowest'
        if (priorityNumber > priorities.size()) {
        	priorityNumber = priorities.size();
        }

        String priority = null;

        for (int i = 0; i < priorityNumber; i++) {
        	priority = ((GenericValue) priorityIt.next()).getString("id");
        }

        return priority;
    }

    
    private String getDefaultSystemPriority() {
        // if priority header is not set, assume it's 'default'
        ConstantsManager constantsManager = ManagerFactory.getConstantsManager();
        GenericValue defaultPriority = constantsManager.getDefaultPriority();
        if (defaultPriority != null) {
            return defaultPriority.getString("id");
        } else {
            log.error("Default priority was null. Using the 'middle' priority.");

            Collection priorities = constantsManager.getPriorities();
            Iterator priorityIt = priorities.iterator();
            int times = (int) Math.ceil((double) priorities.size() / 2d);
            for (int i = 0; i < times; i++) {
                defaultPriority = (GenericValue) priorityIt.next();
            }

            return defaultPriority.getString("id");
        }
    }

    
    /**
     * Given an array of addresses, returns the first valid assignee for the appropriate project.
     * 
     * @param addresses  the addresses
     * @param project  the project
     * @return  the first valid assignee for <code>project</code>
     */
    public static User getFirstValidAssignee(Address[] addresses, GenericValue project)
    {
        if (addresses == null || addresses.length == 0) {
            return null;
        }

        for (int i = 0; i < addresses.length; i++) {
            if (addresses[i] instanceof InternetAddress) {
                InternetAddress email = (InternetAddress) addresses[i];

                try {
                    User validUser = UserUtils.getUserByEmail(email.getAddress());
                    if (isValidAssignee(project, validUser)) {
                    	return validUser;
                    }
                } catch (EntityNotFoundException e) {
                    // loops through the addresses until a valid assignee is found
                }
            }
        }

        return null;
    }


	/**
	 * Tells if <code>user</code> is a valid assignee for <code>project</code>. 
	 * 
	 * @param project  a project
	 * @param user  an user
	 * @return  whether <code>user</code> is a valid assignee for <code>project</code>
	 */
    private static boolean isValidAssignee(GenericValue project, User user) {
    	return (ManagerFactory.getPermissionManager().hasPermission(Permissions.ASSIGNABLE_USER, project, user));	
    }
    
    
    /**
     * Returns an <code>User</code> given its <code>userName</code>.
     * 
     * @param userName  the name of the user
     * @return  the user, or <code>null</code> if there was no user with name <code>userName</code>
     */
    public static User getUserByName(String userName) {
    	try {
			return UserUtils.getUser(userName);
		} catch (EntityNotFoundException e) {
			return null;
		}
    }
    
    
    private void setDefaultSecurityLevel(MutableIssue issue) throws Exception {
        GenericValue project = issue.getProject();
        if (isEnterprise() && project != null) {
            IssueSecurityLevelManager issueSecurityLevelManager = ManagerFactory.getIssueSecurityLevelManager();
            final Long levelId = issueSecurityLevelManager.getSchemeDefaultSecurityLevel(project);
            if (levelId != null) {
                issue.setSecurityLevel(issueSecurityLevelManager.getIssueSecurity(levelId));
            }
        }
    }

    
    private boolean isEnterprise() {
        License license = LicenseManager.getInstance().getLicense(JiraLicenseUtils.JIRA_LICENSE_KEY);
        return (license != null && license.isLicenseLevel(EasyList.build(JiraLicenseUtils.JIRA_ENTERPRISE_LEVEL)));
    }

    
	final void setIssueDescriptor(IssueDescriptor issueDescriptor) {
		this.issueDescriptor = issueDescriptor;
	}
}
