package cern.jira.emailhandler;

import javax.mail.Address;
import javax.mail.Message;
import javax.mail.MessagingException;

import org.apache.log4j.Category;


/**
 * A class containing some useful method to parse a message.
 * <p>Released under the BSD License: see file license.txt for details. 
 * 
 * @author <a href="http://draffo.web.cern.ch">Daniele Raffo</a>
 * @version 2.3 (29-5-2007)
 */
public class MessageParser {

	private static final Category log = Category.getInstance(MessageParser.class);
	
	/** Regex for "any number of characters, even zero". */
	public static final String REGEX_ANYCHAR = ".*";

	
	/**
	 * Parses a message that contains directives about an issue.
	 * 
	 * @param message  the message to parse
	 * @param emailAddresses  an array defining all JIRA email addresses
	 * @return  a descriptor of the issue, following the directives extracted from <code>message</code>
	 */
	public final static IssueDescriptor parse(final Message message, final String[] emailAddresses) {
		IssueDescriptorImpl issue = new IssueDescriptorImpl();
		String toHeader = null;
		try {
			projectkey:
				for (int i = 0; i < emailAddresses.length; i++) {
					toHeader = getRecipientFromMessage(emailAddressToRegex(emailAddresses[i]), message);
					if (toHeader != null) {
						issue.setProjectKey(getFullnameFromHeader(emailAddresses[i], toHeader));
						break projectkey;
					}
				}
			// Project key has been recorded in the issue.
			// Note that at this time the project key may also be invalid, 
			// but may be overridden if a tag Settings.REGEX_PROJECTKEY
			// is found in the message subject.
			if (toHeader == null) {
				// This should never happen as in this case Jira is supposed 
				// not having received the message
				log.error("Jira email address not found amongst recipients.");
				return issue;				
			}
		} catch (MessagingException e) {
			log.error("Error while parsing the message: ", e);
			return issue;
		}
		try {
			parseSubject(message.getSubject(), issue);
		} catch (MessagingException e) {
			log.error("Error parsing message subject: ", e);
			return issue;
		}
		return issue;
	}
	
	
    /**
     * Extracts issue attributes from the email <code>subject</code> and assign them to 
     * <code>issue</code>.
     */
	private final static void parseSubject(final String subject, final IssueDescriptorImpl issue) {
		String summary = new String("");
    	String[] tokens = subject.split("\\s");
    	for (int i = 0; i < tokens.length; i++) {
    		// Check if token matches possible tags.
    		// Issue Type and Priority checks are in increasing importance
    		if (tokens[i].matches(Settings.REGEX_ISSUETYPE_IMPROVEMENT)) {
    			issue.setIssueType("4");        			
    		}
    		else if (tokens[i].matches(Settings.REGEX_ISSUETYPE_SUBTASK)) {
    			issue.setIssueType("5");        			
    		}
    		else if (tokens[i].matches(Settings.REGEX_ISSUETYPE_TASK)) {
    			issue.setIssueType("3");        			
    		}
    		else if (tokens[i].matches(Settings.REGEX_ISSUETYPE_NEW_FEATURE)) {
    			issue.setIssueType("2");        			
    		}
    		else if (tokens[i].matches(Settings.REGEX_ISSUETYPE_BUG)) {
    			issue.setIssueType("1");
    		}
    		else if (tokens[i].matches(Settings.REGEX_PRIORITY_TRIVIAL)) {
    			issue.setPriority(5);        			
    		}
    		else if (tokens[i].matches(Settings.REGEX_PRIORITY_MINOR)) {
    			issue.setPriority(4);        			
    		}
    		else if (tokens[i].matches(Settings.REGEX_PRIORITY_MAJOR)) {
    			issue.setPriority(3);        			
    		}
    		else if (tokens[i].matches(Settings.REGEX_PRIORITY_CRITICAL)) {
    			issue.setPriority(2);        			
    		}
    		else if (tokens[i].matches(Settings.REGEX_PRIORITY_BLOCKER)) {
    			issue.setPriority(1);        			
    		}
    		else if (tokens[i].matches(Settings.REGEX_COMPONENT + MessageParser.REGEX_ANYCHAR)) {
    			String manyComponents = tokens[i].replaceFirst(Settings.REGEX_COMPONENT, "");
    			String[] components = manyComponents.split(",");
    			for (int j = 0; j < components.length; j++) {
    				components[j] = components[j].replaceAll("__", " ");    				
    			}
    	    	issue.setComponents(components);
    		}
    		else if (tokens[i].matches(Settings.REGEX_PROJECTKEY + MessageParser.REGEX_ANYCHAR)) {
    			issue.setProjectKey(tokens[i].replaceFirst(Settings.REGEX_PROJECTKEY, ""));
    			// overrides project key set in recipient email address
    		}
    		else if (tokens[i].matches(Settings.REGEX_REPORTER + MessageParser.REGEX_ANYCHAR)) {
    			issue.setReporter(tokens[i].replaceFirst(Settings.REGEX_REPORTER, ""));
    		}
    		else if (tokens[i].matches(Settings.REGEX_ASSIGNEE + MessageParser.REGEX_ANYCHAR)) {
    			issue.setAssignee(tokens[i].replaceFirst(Settings.REGEX_ASSIGNEE, ""));
    		}
    		else {   // Token is not a tag, then add it to the summary text 
    			summary += (tokens[i] + " ");
    		}
    	}
    	summary = summary.trim();
    	issue.setSummary(summary.equals("") ? Settings.DEFAULT_SUMMARY : summary);
    }

    
    /**
     * Retrieves the To:, CCc:, or Bcc: header purporting to a given recipient of a email message.
     * 
     * @param recipient  a regex describing the recipient email address  
     * @param message  an email message
     * @return  the header of <code>message</code> containing <code>recipient</code>, 
     * or <code>null</code> if the <code>recipient</code> was not found 
     * @throws MessagingException  if there were problems handling the message
     */
    public final static String getRecipientFromMessage(final String recipient, final Message message) throws MessagingException {
    	Address[] addresses = message.getAllRecipients();
    	for (int i = 0; i < addresses.length; i++) {
    		log.debug("Parsing message address " + i + ": " + addresses[i]);
    		if (addresses[i].toString().matches(recipient)) {
    			return addresses[i].toString(); 
    		}
    	}
    	return null;
    }
    
    
    /**
     * Extract the "Full Name" part from a full email address (which could be
     * contained in a From:, To:, Cc:, or Bcc: header). 
     * E.g. if <code>fullEmail</code> is "Arthur Dent &lt;arthur@vogon.org&gt;" and 
     * <code>emailAddress</code> is "arthur@vogon\\.org", it returns "Arthur Dent".
     * This method correctly manages the fact that the emailclient could surround 
     * email address and/or full name with characters like <code><> " ' () </code>, etc. 
     * 
     * @param emailAddress  a regex specifying the bare email address to extract
     * @param fullEmail  the full email address
     * @return  the Full Name
     */
    public final static String getFullnameFromHeader(final String emailAddress, final String fullEmail) {
    	String[] tokensToDelete = {emailAddress, "<", ">", "\"", "\'", "\\(", "\\)"};
    	String fullName = new String(fullEmail);
    	for (int i = 0; i < tokensToDelete.length; i++) {
    		fullName = fullName.replaceAll(tokensToDelete[i], "");
    	}
    	fullName = fullName.trim();
		log.debug("Extracted from header = " + fullEmail + " the fullName = " + fullName);
    	return fullName;
    }
    
    
    /**
     * Transforms an email address into a regular expression matching that address.
     * 
     * @param emailAddress  an email address of kind <code>user.login@domain.xx</code> 
     * @return  a regex matching <code>emailAddress</code>
     */
    public final static String emailAddressToRegex(final String emailAddress) {
    	StringBuffer regex = new StringBuffer(".*");
    	for (int c = 0; c < emailAddress.length(); c++) {
    		if (emailAddress.charAt(c) == '.') {
    			regex.append("\\.");
    		} else {
    			regex.append(emailAddress.charAt(c));
    		}
    	}
    	return (regex + ".*").toString();
    }

    
}

